/*
 * This is the ChaCha20 quarter round. For it to work, it takes four indices from the _ks matrix,
 * and assumes the existence of four /pif_LongInt/Unsigned32 objects which have already been
 * initialized and labeled as a, b, c, and d.
 *
 * Note: The key and nonce are expected to be strings, respectively, of 64 and 24 hex digits.
 */

#define	pCC_ChaCha20_QR(i1, i2, i3, i4) \
	a.Set(_ks[(i1)+1], _ks[(i1)]); \
	b.Set(_ks[(i2)+1], _ks[(i2)]); \
	c.Set(_ks[(i3)+1], _ks[(i3)]); \
	d.Set(_ks[(i4)+1], _ks[(i4)]); \
	a += b; d ^= a; d.BitshiftLeftRotate(16); \
	c += d; b ^= c; b.BitshiftLeftRotate(12); \
	a += b; d ^= a; d.BitshiftLeftRotate( 8); \
	c += d; b ^= c; b.BitshiftLeftRotate( 7); \
	_ks[(i1)+1] = a.block_2; _ks[(i1)] = a.block_1; \
	_ks[(i2)+1] = b.block_2; _ks[(i2)] = b.block_1; \
	_ks[(i3)+1] = c.block_2; _ks[(i3)] = c.block_1; \
	_ks[(i4)+1] = d.block_2; _ks[(i4)] = d.block_1;

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_CRYPTOSUITE)
pif_CryptoSuite/ChaCha20
#else
CryptoSuite/ChaCha20
#endif
	/*
	 * Variables.
	 */

	var
		// Stores the plaintext of the key and nonce, for reading.
		_keyPlain
		_noncePlain

		list
			// In Encrypt() this is used to convert a Unicode string into a sequence of UTF-8 bytes,
			// and in Decrypt() this does this reverse before converting the bytes into a string.
			_bitstream

			// Stores the key and nonce as a list of numbers in ASCII format.
			_key
			_nonce

			// The matrix state of the algorithm, which is used for encryption and decryption.
			_state

		// Stores the counter state. _counter2 is more-significant, and _counter1 is less-
		// significant.
		_counter2 = 0
		_counter1 = 0

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_LONGINT)
		pif_LongInt/Unsigned32
#else
		LongInt/Unsigned32
#endif
			u32

			// Used for the quarter rounds in the key stream.
			a
			b
			c
			d

		const
			// There are actually 20 rounds, but we will perform 2 rounds each iteration.
			ROUNDS = 10

			// The normal ChaCha20 algorithm uses a 16 block matrix for the state/ but we have to
			// store each 32 bit integer in two positions, so instead of being 4x4 it's 4x8 (well,
			// it "looks" like it's 4x8, but it's still as far as the math is concerned 4x4).
			STATE = 32

			// The size of a block of text. This is also how many bytes the keystream is.
			BLOCK_SIZE = 64

			// In bytes
			KEY_LENGTH   = 32
			NONCE_LENGTH = 12 // In this variant, the counter is only 32 bit, and the nonce gets the
							  // extra eight bits.

			// This is how many 'blocks' (list indices) the key and nonce resp. will use. Because
			// each block corresponds to two bytes, it's just the values above divided by two.
			KEY_BLOCK    = KEY_LENGTH / 2
			NONCE_BLOCK  = NONCE_LENGTH / 2

			// Corresponds to the string "expand 32-byte k". This is used to set up the
			// state.
			CONST_41 = 0x7865
			CONST_42 = 0x6170
			CONST_31 = 0x646E
			CONST_32 = 0x3320
			CONST_21 = 0x2D32
			CONST_22 = 0x7962
			CONST_11 = 0x6574
			CONST_12 = 0x6B20

	/*
	 * Constructor.
	 */

	New(Key = null, Nonce = null, Counter = 0)
		if(Key != null)
			SetKey(Key)
		if(Nonce != null)
			SetNonce(Nonce)
		if(Counter !=  0)
			SetCounter(Counter)

		return 1

	/*
	 * Implementation of pif_CryptoSuite methods.
	 */

	SetKey(Key)
		if(length(Key) == 2*KEY_LENGTH)
			// 2*KEY_LENGTH becase we're reading in hex values.

			_keyPlain = Key

			if(!_key)
				_key = new(KEY_BLOCK)

			var
				// The block for the characters.
				b

				// Stores a particular character for processing as we convert
				// it into a hex value.
				c
			for(var/i = 1, i <= 2*KEY_LENGTH, i += 4)
				// We work 4 characters at a time, which amounts to two bytes.

				b = 0

				c = text2ascii(Key[i])
				c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48) // Quick conversion from decimal to
				b |= (c << 4)								 // hexadecimal.

				c = text2ascii(Key[i+1])
				c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
				b |= c

				c = text2ascii(Key[i+2])
				c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
				b |= (c << 12)

				c = text2ascii(Key[i+3])
				c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
				b |= (c << 8)

				_key[((i - 1) / 4) + 1] = b

			return 1

		return 0

	Key()
		return _keyPlain

	Encrypt(Plaintext, Key = null, Nonce = null, Counter = null)
		// This takes the plaintext, along with an optional key, nonce, and counter (key and nonce
		// only optional if they have already been set), and outputs a hexadecimal-encoded
		// ciphertext.

		if(Key != null)
			SetKey(Key)
		if(Nonce != null)
			SetNonce(Nonce)
		if(Counter != null)
			SetCounter(Counter)

		_init()

		// Initialize these in preparation for computing the keystreams.
		a = new
		b = new
		c = new
		d = new

		// Convert the plaintext into a sequence of UTF8 bytes, which we'll encrypt.
		_utf8encode(Plaintext)

		var
			// Number of bytes in the bistream.
			bytes = _bitstream.len

			// Points to the current position in _bitstream
			pos = 0

			// Stores the bytes of the bitstream in their proper arrangement, just before being
			// XORed with the keystream. We store these because they need to be XORed with the
			// keystream slightly out of the order we'll grab them if we need to get more than 2
			// bytes of the bitstream.
			char1
			char2

			// The number of blocks from the keystream we'll be dealing with. At minimum there will
			// always be at least 1.
			blocks = -round(-bytes / BLOCK_SIZE) || 1

			// Use when computing the new state of the matrix during the rounds block of the loop.
			// The "ks" stands for "keystate".
			list/_ks = new(STATE)

			hex // Stores the hex after XORing the _bitstream with the keystream.
			Ciphertext // The output. We get this by XORing the output from the keystream with each
			           // utf8 byte from _bitstream.

		for(var/i = 1 to blocks)
			// Each block can be used to process 16 32-bit integers from _bitstream. Because
			// _bitstream is broken into (obviously 8-bit) bytes, this involves some finangling
			// down below (twss &c).

			_keystream(_ks) // Pass the keystream to the _keystream() method, which will modify it
			                // in place. The value will then be stored in the keystream.

			// We increment the counter after every computing every keystream, per the ChaCha20
			// algorithm.
			Increment()

			for(var/k = 1, k <= STATE, k += 2)
				// We add 2 at each step because, as discussed much further above, _keystream
				// consists of a pair of 16-bit words glued together via an Unsigned32 LongInt
				// object into one 32-bit integer.

				// Store the current state of the keystream.
				a.Set(_ks[k+1], _ks[k])

				switch(bytes - pos)
					// We do this to test when we're getting close to the end of _bitstream, so we
					// can figure out exactly when to terminate the loops.

					if(0)
						// We've reached the actual end of the _bitstream, so we can exit the loops.
						// We use a goto here, in a rare justified example, so we don't have to work
						// with awkward bitflags to escape the outer loop. It's a lot cleaner this
						// way.

						goto EXIT_LOOP

					if(1)
						// If there's only one byte left, we'll only bother with the first byte of
						// the keystream.

						a.BitwiseXor(
							// XOR the keystream with the relevant byte of the _bitstream.
							0,
							_bitstream[++pos]
						)
						hex = a.PrintHexadecimal()

						// Concatenate in the right order. a.PrintHexadecimal() will output a value
						// prepended with 0x, so we also want to skip over that part.
						Ciphertext = Ciphertext + hex[7] + hex[8]

						// No more bytes, so exit the loop.
						goto EXIT_LOOP

					if(2)
						// Same as above, but instead we look at the first two bytes of the
						// bitstream.

						a.BitwiseXor(
							0,
							_bitstream[++pos] | (_bitstream[++pos] << 8)
						)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6]

						goto EXIT_LOOP

					if(3)
						// Three bytes of the bitstream.

						char1 = _bitstream[++pos] | (_bitstream[++pos] << 8)
						char2 = _bitstream[++pos]

						a.BitwiseXor(char2, char1)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6] + \
							hex[3] + hex[4]

						goto EXIT_LOOP

					else
						// If none of the above is satisfied, we XOR all the current bytes of the
						// keystream with the bitstream. After this, we continue as there may be
						// more left in the bitstream.

						char1 = _bitstream[++pos] | (_bitstream[++pos] << 8)
						char2 = _bitstream[++pos] | (_bitstream[++pos] << 8)

						a.BitwiseXor(char2, char1)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6] + \
							hex[3] + hex[4] + \
							hex[1] + hex[2]

		// Used to terminate the loop when there are no characters left in the bitstream.
		EXIT_LOOP

		// Clear the Unsigned32 objects so they can be GC'd.
		a = null
		b = null
		c = null
		d = null

		// Also clear our _bitstream, so it's empty for the next time we do anything.
		_bitstream.len = 0

		// Lastly, return the final ciphertext.
		return Ciphertext

/*
 * Old code for Encrypt(), from before Unicode support. Valid if string is ASCII only.
 *

	Encrypt(Plaintext, Key = null, Nonce = null, Counter = null)
		// This takes the plaintext, along with an optional key, nonce, and counter (key and nonce
		// only optional if they have already been set), and outputs a hexadecimal-encoded
		// ciphertext.

		if(Key)
			SetKey(Key)
		if(Nonce)
			SetNonce(Nonce)
		if(Counter != null)
			// Counter == 0 is valid
			SetCounter(Counter)

		_init()

		// Initialize these in preparation for computing the keystreams.
		a = new
		b = new
		c = new
		d = new

		var
			blocks = -round(-length(Plaintext) / BLOCK_SIZE) || 1

			// Using when computing the new state of the matrix during the rounds block of the loop.
			// The "ks" stands for "keystate"
			list/_ks = new(STATE)

			pos = 0 // Points to our current position in the plaintext.
			Ciphertext // The output. We get this by XORing the output from the keystream with each
			           // character from the plaintext.
			hex // Stores the hex after XORing the keystream with the plaintext.

			// Stores the characters from the plaintext, just prior to being XORed with the
			// keystream.
			char1
			char2

			// Length of the plaintext.
			pLen = length(Plaintext)

		for(var/i = 1 to blocks)
			_keystream(_ks) // Pass the keystream to the _keystream() method, which will modify it
							// in place. The value will then be stored in the keystream.

			// We increment the counter after every computing every keystream.
			Increment()

			for(var/k = 1, k <= STATE, k += 2)
				a.Set(_ks[k+1], _ks[k]) // Store the current element of the keystream.

				switch(pLen - pos)
					// This is mostly to test for when we're getting close to the end of the
					// plaintext. In those cases, we have to do things a little bit differently.

					if(0)
						// If there's no bytes left in the keystream, just terminate immediately.
						goto EXIT_LOOP

					if(1)
						// If there's only one character, we only bother with the first byte of the
						// keystream.

						a.BitwiseXor(
							// XOR the keystream with the plaintext characters.
							0,
							text2ascii(Plaintext, ++pos)
						)
						hex = a.PrintHexadecimal()

						// Concatenate.
						Ciphertext = Ciphertext + hex[7] + hex[8]

						// There's no more characters left, so terminate the loop.
						goto EXIT_LOOP

					if(2)
						// Same as above, but instead we look at the first two bytes of the
						// keystream.

						a.BitwiseXor(
							0,
							text2ascii(Plaintext, ++pos) | (text2ascii(Plaintext, ++pos) << 8)
						)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6]

						goto EXIT_LOOP

					if(3)
						// Three btyes of the keystream.

						char1 = text2ascii(Plaintext, ++pos) | (text2ascii(Plaintext, ++pos) << 8)
						char2 = text2ascii(Plaintext, ++pos)

						a.BitwiseXor(char2, char1)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6] + \
							hex[3] + hex[4]

						goto EXIT_LOOP

					else
						// If none of the above is satisfied, we XOR all the current bytes of the
						// keystream with the plaintext. After this, we continue as there may be
						// more left in the plaintext.

						char1 = text2ascii(Plaintext, ++pos) | (text2ascii(Plaintext, ++pos) << 8)
						char2 = text2ascii(Plaintext, ++pos) | (text2ascii(Plaintext, ++pos) << 8)

						a.BitwiseXor(char2, char1)
						hex = a.PrintHexadecimal()

						Ciphertext = Ciphertext + \
							hex[7] + hex[8] + \
							hex[5] + hex[6] + \
							hex[3] + hex[4] + \
							hex[1] + hex[2]

			// Used to terminate the loop when there are no characters left in the plaintext.
			EXIT_LOOP

		// Clear the Unsigned32 objects so they can be GC'd.
		a = null
		b = null
		c = null
		d = null

		return Ciphertext

 *
 * End pre-Unicode code for Encrypt()
 */

	Decrypt(Ciphertext, Key = null, Nonce = null, Counter = null)
		// This takes a hexadecimal-encoded ciphertext, along with an optional key, nonce, and
		// counter. As with Encrypt(), the key and nonce are only optional if they haven't been set
		// already. The output is the Unicode plaintext.

		if(Key != null)
			SetKey(Key)
		if(Nonce != null)
			SetNonce(Nonce)
		if(Counter != null)
			SetCounter(Counter)

		// For notes on the set up, look at Encrypt(). Much of it is similar to here.

		_init()

		a = new
		b = new
		c = new
		d = new

		if(!_bitstream)
			// If the _bitstream list has not be instantiated, do so. We'll be reading utf8 data
			// into it before converting back to Unicode.
			_bitstream = new
		else
			// If it is instantiated, then make sure there's nothing in it.
			_bitstream.len = 0

		var
			// The number of 128-character blocks in the Ciphertext. Instead of looking at 64-char
			// blocks like with Encrypt(), we double the size because two hexadecimal digits store
			// one byte.
			blocks = -round(-length(Ciphertext) / (2*BLOCK_SIZE)) || 1

			list/_ks = new(STATE) // Stores the keystream.

			pos = 0 // Points to the current position in the ciphertext.
			cLen = length(Ciphertext)
			cHex // Hex for the characters in the ciphertext.

			// Plaintext

		for(var/i = 1 to blocks)
			// For each block, grab the corresponding keystream and increment the counter.

			_keystream(_ks)
			Increment()

			for(var/k = 1, k <= STATE, k += 2)
				a.Set(_ks[k+1], _ks[k]) // Element of the keystream.

				switch(cLen - pos)
					if(0)
						// If there's no characters left, terminate the loop.
						goto EXIT_LOOP

					if(2)
						cHex = Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						_bitstream += b.block_1 & 0x00FF

						goto EXIT_LOOP

					if(4)
						cHex = Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						_bitstream.Add(
							 b.block_1 & 0x00FF,
							(b.block_1 & 0xFF00) >> 8
						)

						goto EXIT_LOOP

					if(6)
						cHex = Ciphertext[pos+5] + Ciphertext[pos+6] + \
						       Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						_bitstream.Add(
							 b.block_1 & 0x00FF,
							(b.block_1 & 0xFF00) >> 8,
							 b.block_2 & 0x00FF
						)

						goto EXIT_LOOP

					else
						cHex = Ciphertext[pos+7] + Ciphertext[pos+8] + \
						       Ciphertext[pos+5] + Ciphertext[pos+6] + \
						       Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						_bitstream.Add(
							 b.block_1 & 0x00FF,
							(b.block_1 & 0xFF00) >> 8,
							 b.block_2 & 0x00FF,
							(b.block_2 & 0xFF00) >> 8
						)

						pos += 8

			EXIT_LOOP // Used to escape the above loop if we "run out of" ciphertext.

		a = null
		b = null
		c = null
		d = null

		// Return the decoded Unicode string.
		. = _utf8decode()

		// Clear out the _bitstream for the next time we use it.
		_bitstream.len = 0

/*
 * Old code for Decrypt(), from before Unicode support. Valid if string is ASCII only.
 *

	Decrypt(Ciphertext, Key = null, Nonce = null, Counter = null)
		// This takes a hexadecimal-encoded ciphertext, along with an optional key, nonce, and
		// counter. As with Encrypt(), the key and nonce are only optional if they haven't been
		// set already. The output is the ASCII plaintext.

		if(Key)
			SetKey(Key)
		if(Nonce)
			SetNonce(Nonce)
		if(Counter != null)
			// Counter = 0 is valid
			SetCounter(Counter)

		// For notes on the set up, look at Encrypt(). Much of it is similar to here.

		_init()

		a = new
		b = new
		c = new
		d = new

		var
			// The number of 128-character blocks in the Ciphertext. Instead of looking at 64-char
			// blocks like with Encrypt(), we double the size because two hexadecimal digits store
			// one byte.
			blocks = -round(-length(Ciphertext) / (2*BLOCK_SIZE)) || 1

			list/_ks = new(STATE) // Stores the keystream.


			pos = 0 // Points to the current position in the ciphertext.
			cLen = length(Ciphertext)
			cHex // Hex for the characters in the ciphertext.

			Plaintext

		for(var/i = 1 to blocks)
			// For each block, grab the corresponding keystream and increment the counter.

			_keystream(_ks)
			Increment()

			for(var/k = 1, k <= STATE, k += 2)
				a.Set(_ks[k+1], _ks[k]) // Element of the keystream.

				switch(cLen - pos)
					if(0)
						// If there's no characters left, terminate the loop.
						goto EXIT_LOOP

					if(2)
						cHex = Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						Plaintext += ascii2text(b.block_1 & 0x00FF)

						goto EXIT_LOOP

					if(4)
						cHex = Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						Plaintext += ascii2text( b.block_1 & 0x00FF)
						Plaintext += ascii2text((b.block_1 & 0xFF00) >> 8)

						goto EXIT_LOOP

					if(6)
						cHex = Ciphertext[pos+5] + Ciphertext[pos+6] + \
						       Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						Plaintext += ascii2text( b.block_1 & 0x00FF)
						Plaintext += ascii2text((b.block_1 & 0xFF00) >> 8)
						Plaintext += ascii2text( b.block_2 & 0x00FF)

						goto EXIT_LOOP

					else
						cHex = Ciphertext[pos+7] + Ciphertext[pos+8] + \
						       Ciphertext[pos+5] + Ciphertext[pos+6] + \
						       Ciphertext[pos+3] + Ciphertext[pos+4] + \
						       Ciphertext[pos+1] + Ciphertext[pos+2]

						b.Set("0x" + cHex)
						b ^= a

						Plaintext += ascii2text( b.block_1 & 0x00FF)
						Plaintext += ascii2text((b.block_1 & 0xFF00) >> 8)
						Plaintext += ascii2text( b.block_2 & 0x00FF)
						Plaintext += ascii2text((b.block_2 & 0xFF00) >> 8)

						pos += 8

			EXIT_LOOP // Used to escape the above loop if we "run out of" ciphertext.

		a = null
		b = null
		c = null
		d = null

		return Plaintext

 *
 * End pre-Unicode code for Decrypt()
 */

	proc
		/*
		 * "Public" methods.
		 */

		SetNonce(Nonce)
			if(length(Nonce) == 2*NONCE_LENGTH)
				_noncePlain = Nonce

				if(!_nonce)
					_nonce = new(NONCE_BLOCK)

				var
					// The block for the characters.
					b

					// Stores a particular character for processing as we convert
					// it into a hex value.
					c
				for(var/i = 1, i <= 2*NONCE_LENGTH, i += 4)
					// We work 4 characters at a time, which amounts to two bytes.

					// n = (n > 57) ? ( (n & 32) - 54 ) : (n - 48)

					b = 0

					c = text2ascii(Nonce[i])
					c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
					b |= (c << 4)

					c = text2ascii(Nonce[i+1])
					c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
					b |= c

					c = text2ascii(Nonce[i+2])
					c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
					b |= (c << 12)

					c = text2ascii(Nonce[i+3])
					c = (c > 57) ? ( (c & ~32) - 55 ) : (c - 48)
					b |= (c << 8)

					_nonce[((i - 1) / 4) + 1] = b

				return 1

			return 0

		Nonce()
			return _noncePlain

		Counter()
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_LONGINT)
			return new /pif_LongInt/Unsigned32(_counter2, _counter1)
#else
			return new /LongInt/Unsigned32(_counter2, _counter1)
#endif

		SetCounter(...)
			// Sets the counter using a GAAF format.

			if(!u32)
				u32 = new(args)
			else
				u32.Set(args)

			_counter2 = u32.block_2
			_counter1 = u32.block_1

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_LONGINT)
			return new /pif_LongInt/Unsigned32(u32)
#else
			return new /LongInt/Unsigned32(u32)
#endif

		Increment()
			if(!u32)
				u32 = new(_counter2, _counter1)
			else
				u32.Set(_counter2, _counter1)

			u32.Increment()
			_counter2 = u32.block_2
			_counter1 = u32.block_1

			return 1

		/*
		 * "Private" methods.
		 */

		_utf8encode(_string)
			// This takes a string and converts it into UTF-8 bytes, which are stored in the
			// _bistream list.

			if(!_bitstream)
				// If the _bistream list hasn't been created before, then do so.
				_bitstream = new
			else
				// Otherwise, reset its length to 0 and we'll just reuse the object.
				_bitstream.len = 0

			var
				// Number of bytes in the string.
				bytes = length(_string)

				// The Unicode codepoint of a given character.
				codepoint

			for(var/i = 1 to bytes)
				codepoint = text2ascii(_string, i)

				switch(codepoint)
					// UTF-8 encodes each Unicode character differently depending on the codepoint.
					// The main effect is the number of bytes used to encode the character, from one
					// to four bytes.

					if(0x000000 to 0x00007F) // 0 to 127
						// codepoint: U+(0abc defg)
						// utf8:      0abcdefg

						// bitstream += (0x00007F & codepoint)
						_bitstream += codepoint

					if(0x000080 to 0x0007FF) // 128 to 2047
						// codepoint: U+(0000 0abc defg hijk)
						// utf8:      110abcde 10fghijk

						_bitstream.Add(
							0x0000C0 | ((0x0007C0 & codepoint) >> 6),
							0x000080 |  (0x0000FF & codepoint)
						)

						// This is a two byte character, so we bump the counter up by one.
						i ++

					if(0x000800 to 0x00FFFF) // 2048 to 65536
						// codepoint: U+(abcd efgh ijkl mnop)
						// utf8:      1110abcd 10efghij 10klmnop

						_bitstream.Add(
							0x0000E0 | ((0x00F000 & codepoint) >> 12),
							0x000080 | ((0x000FC0 & codepoint) >>  6),
							0x000080 |  (0x00003F & codepoint)
						)

						// Three byte, so bump up by two.
						i += 2

					if(0x010000 to 0x1FFFFF)
						// codepoint: U+(000a bcde fghi jklm nopq rstu)
						// utf8:      11110abc 10defghi 10jklmno 10pqrstu

						_bitstream.Add(
							0x0000F0 | ((codepoint & 0x1C0000) >> 16),
							0x000080 | ((codepoint & 0x03F000) >> 12),
							0x000080 | ((codepoint & 0x000FC0) >>  6),
							0x000080 |  (codepoint & 0x00003F)
						)

						// Four byte, so bump up by 3.
						i += 3

			// Now that we've converted to utf8 successfully, pass the _bitstream list back to the
			// calling proc.
			return _bitstream

		_utf8decode()
			// Reads from _bistream to decode from UTF-8 back into a Unicode string.

			// Prep the return string.
			. = ""

			var
				// The total number of bytes in the bitstream.
				bytes = _bitstream.len

				// The current utf8 byte we're looking at.
				utf8_byte

				// The unicode character we're reconstruction.
				unicode_char

			for(var/i = 1, i <= bytes, i ++)
				// This *should* only be reading leading bytes, and then we'll individually process
				// the continuation bytes and adjust the counter i manually to move to the right
				// position.

				utf8_byte = _bitstream[i]

				if( (utf8_byte & 0x000080) == 0x000000)
					// This is a one-byte UTF-8 encoding, so we can write it directly to the string.

					// codepoint: U+(0abc defg)
					// utf8:      0abcdefg
					unicode_char = utf8_byte

				else if( (utf8_byte & 0x0000E0) == 0x0000C0)
					// This leading byte is for a two-byte UTF-8 encoding.

					// codepoint: U+(0000 0abc defg hijk)
					// utf8:      110abcde 10fghijk
					unicode_char  = (utf8_byte       & 0x0001F) << 6
					unicode_char |=  _bitstream[++i] & 0x0003F

				else if( (utf8_byte & 0x0000F0) == 0x0000E0)
					// This leading byte is for a three-byte UTF-8 encoding.

					// codepoint: U+(abcd efgh ijkl mnop)
					// utf8:      1110abcd 10efghij 10klmnop
					unicode_char  = (utf8_byte       & 0x00000F) << 12
					unicode_char |= (_bitstream[++i] & 0x00003F) <<  6
					unicode_char |=  _bitstream[++i] & 0x00003F

				else if( (utf8_byte & 0x0000F8) == 0x0000F0)
					// Lastly, this leading byte is for a four-byte UTF-8 encoding.

					// codepoint: U+(000a bcde fghi jklm nopq rstu)
					// utf8:      11110abc 10defghi 10jklmno 10pqrstu
					unicode_char  = (utf8_byte       & 0x000007) << 18
					unicode_char |= (_bitstream[++i] & 0x00003F) << 12
					unicode_char |= (_bitstream[++i] & 0x00003F) <<  6
					unicode_char |=  _bitstream[++i] & 0x00003F

				// Now that we have reconstructed the unicode character, write it to the string.
				. += ascii2text(unicode_char)

		_init()
			// Initializes the state from the nonce, key, constant, and counter.
			_state = new(STATE - 2) // Subtract 2 to account for the counter not being
									// (see below)

			/*

			The following diagram is how the "virtual" matrix will look, with number corrsponding to
			list indices.

			[ 1  3  5  7]    [ C  C  C  C]    - C is the constant
			[ 9 11 13 15]    [ K  K  K  K]    - K is the key
			[17 19 21 23] <= [ K  K  K  K]    - P is the counter
			[ P 25 27 29]    [ P  N  N  N]    - N is the nonce

			P on the left corresponds to the counter, which has its own dedicated variables. Because
			of this, we won't actually be storing it in the state.

			*/

			// Blocks for the constant.
			_state[ 1] = CONST_41
			_state[ 2] = CONST_42

			_state[ 3] = CONST_31
			_state[ 4] = CONST_32

			_state[ 5] = CONST_21
			_state[ 6] = CONST_22

			_state[ 7] = CONST_11
			_state[ 8] = CONST_12

			// Blocks for the key.
			for(var/i = 1 to KEY_BLOCK)
				_state[8+i] = _key[i]

			// Blocks for the nonce.
			for(var/i = 1 to NONCE_BLOCK)
				_state[24+i] = _nonce[i]

		_keystream(_ks)
			// This method generates the keystream (though we don't bother with serializing the
			// keystream) by writing it to the _ks list specified.

			for(var/k = 1 to (STATE / 2))
				// Transfer the shit from the _state matrix to the _ks matrix.

				var/q,r
				if(k < 13)
					q = _state[(2*k)-1]
					r = _state[ 2*k   ]
				else if(k == 13)
					q = _counter1
					r = _counter2
				else
					q = _state[(2*k)-3]
					r = _state[(2*k)-2]

				_ks[(2*k)-1] = q || 0
				_ks[ 2*k   ] = r || 0

			for(var/j = 1 to ROUNDS)
				/*
				 * This is where the actual meat of the encryption algorithm occurs, where we
				 * perform the quarter rounds on the rows and diagonals of the state matrix. The
				 * quarter rounds are as follows, where += denotes addition mod 2^32, ^= is XOR, and
				 * <<<= is bitshift-left-rotate.
				 *
				 * 		QR(a,b,c,d):
				 * 			a += b; d ^= a; d <<<= 16
				 *			c += d; b ^= c; b <<<= 12
				 *			a += b; d ^= a; d <<<= 8
				 *			c += d; b ^= c; b <<<= 7
				 *
				 * Each iteration of the loop goes through two rounds, which is known as a double
				 * round. The first round modifies the columns of the state, and the second route
				 * modifies the diagonals. Below, we will denote the "virtual" quarter round by QR
				 * and the "actual" quarter round by qr. The "virtual" quarter round is based on how
				 * the matrix looks in theory, i.e. how it's specified in the algorithm, while the
				 * "actual" quarter round is based on how it's represented in our system.
				 *
				 * Here we do the odd round, i.e. the first four quarter rounds acting on the
				 * columns of the state.
				 */

				pCC_ChaCha20_QR(1,  9, 17, 25) // QR(0, 4,  8, 12) = qr(1,  9, 17, 25)
				pCC_ChaCha20_QR(3, 11, 19, 27) // QR(1, 5,  9, 13) = qr(3, 11, 19, 27)
				pCC_ChaCha20_QR(5, 13, 21, 29) // QR(2, 6, 10, 14) = qr(5, 13, 21, 29)
				pCC_ChaCha20_QR(7, 15, 23, 31) // QR(3, 7, 11, 15) = qr(7, 15, 23, 31)

				/*
				 * How we do the even round, i.e. the second four quarter rounds which act on the
				 * diagonals of the state.
				 */

				pCC_ChaCha20_QR(1, 11, 21, 31) // QR(0, 5, 10, 15) = qr(1, 11, 21, 31)
				pCC_ChaCha20_QR(3, 13, 23, 25) // QR(1, 6, 11, 12) = qr(3, 13, 23, 25)
				pCC_ChaCha20_QR(5, 15, 17, 27) // QR(2, 7,  8, 13) = qr(5, 15, 17, 27)
				pCC_ChaCha20_QR(7,  9, 19, 29) // QR(3, 4,  9, 14) = qr(7,  9, 19, 29)

			for(var/k = 1, k <= STATE, k += 2)
				// Now that we've done the rounds, we add the individual values to the actual _state
				// matrix.

				a.Set(_ks[k+1], _ks[k])

				if(k < 25)
					// Before we get to the counter, we can pull directly from the _state list.
					a.Add(_state[k+1], _state[k])
				else if(k == 25)
					// In this case ,we're modifying the counter.
					a.Add(_counter2, _counter1)
				else
					// After it, we need to subtract 2 from the index, since we "skipped over" the
					// counter in the list.
					a.Add(_state[k-1], _state[k-2])

				_ks[k  ] = a.block_1
				_ks[k+1] = a.block_2

#undef pCC_ChaCha20_QR